# Moritz Helias, 2014

# Correlations in LIF or Hawkes networks
# cf. Grytskyy et al. Front Comput Neurosci (2013)

import numpy as np
import scipy as sp
import scipy.special as sps
import mpmath as mpf


class corr_hawkes_like:

    def __init__(self, tau, d, gamma, g, epsilon, NE, w, nu, sigma_a, alpha, poles_max):
        self.tau = tau # time constant
        self.d = d     # delay
        self.epsilon = epsilon
        self.gamma = gamma
        self.g = g
        self.w = w
        self.NE = NE
        self.nu = nu
        self.sigma_a = sigma_a
        self.alpha = alpha
        self.poles_max = poles_max

    def H_(self, z):
        return 1./(1.+self.tau*z)*np.exp(-z*self.d)

        
    def get_poles_(self, L, branch_max):
        # Calculate poles of recurrent transfer function
        # using the Lambert W function. branch_max specifies the
        # positive maximum branch until which to go. The first two
        # entries are the solutions from the principal branch, which are
        # real for z > -1/e and which are conjugate pairs for z < -1/e.
        #
        #K = self.gamma*(self.CE - self.g*self.CI)

        if self.d == 0.0:
            return [(L-1)/self.tau]
        
        z0s = []
        for branch in range(-1, branch_max):
            if L != 0.0:
                z0 = -1.0/self.tau + 1.0 / self.d * mpf.lambertw(L*self.d/self.tau*np.exp(self.d/self.tau), branch)
                # check: should be zero
                z0 = complex(z0)
                zr = 1.0 - L*self.H_(z0)
                assert(abs(zr) < 1e-10)
                # if assertion does not break, it works 
            else:
                z0 = -1.0/self.tau
                if branch > -1:
                    break

            z0s.append(z0)

            # add missing conjugate complex poles for branch >= 1            
            if branch >= 1:
               z0s.append(z0.conjugate())

        return z0s


    def Gp(self, p, x, sigma):
        """Convolution of exponential function and Gaussian expressed by Faddeev function"""
        return sigma * np.sqrt(np.pi/2.) * sps.wofz(complex(0., 1./np.sqrt(2.)*(sigma*p-x/sigma)))*np.exp(p*x-x**2/(2.*sigma**2))

    def f_1_star_a(self, p, t, d, sigma):
        """Function f1 convolved with Gaussian"""
        return (self.Gp(-p, t-d, sigma)-self.Gp(-p, -d, sigma)) * np.exp(p*(t-d))

    def r(self, p, t, d, sigma):
        """Helper finction for f2"""
        return (-d < t)*(t < d)*np.exp(p*t)*(self.Gp(-p, t, sigma)-self.Gp(-p, -d, sigma))\
               + (t>=d)*np.exp(p*t)*(self.Gp(-p, d, sigma)-self.Gp(-p, -d, sigma))

    def f_2_star_a(self, p, t, d, sigma):
        """Function f2 convolved with Gaussian"""
        return self.r(p, t, d, sigma) + self.r(p, -t, d, sigma)

    def eta(self, sigma):
        return np.exp(self.d/self.tau)*self.Gp(1./self.tau, -self.d, sigma)


    def cross_cov_matrix(self):
        """Calculates the cross covariance matrix for positive t"""

        # mean rate feedback
        L = self.epsilon*self.NE*self.w*(1. - self.gamma*self.g)

        print 'L = ', L

        # poles of resonant denominator
        pks = self.get_poles_(L, self.poles_max)
    
        m1_mat = self.epsilon * self.w * np.matrix([ [1.0, -self.g], [1.0, -self.g] ])

        m3_mat = self.epsilon**2 * self.NE * self.w**2 * (1. + self.gamma*self.g**2) * np.matrix( [ [1.0, 1.0], [1.0, 1.0] ] )
        
        def F1k(p):
            return 1./((1. + p*self.tau)*self.d + self.tau)

        def F2k_div_F1k(p):
            return 1. / (1. - p*self.tau - L*np.exp(p*self.d))

        eta = self.eta(self.sigma_a)
        
        def c(t_prime):
            # use symmetry: c^T(t) = c(-t) to calculate for negative t
            if t_prime < 0:
                t = -t_prime
            else:
                t = t_prime


            # all poles are returned, conjugate pairs are in the list explicitly
            f1_delta = 0.0
            f2_delta = 0.0
            f1_a = 0.0
            f2_a = 0.0
            
            for pk in pks:
                E = np.exp(pk*t)
                E_m_d = np.exp(pk*abs(t-self.d))
                E_p_d = np.exp(pk*(t+self.d))
                F1p = F1k(pk)
                F2p = F1p * F2k_div_F1k(pk)

                # contributions due to delta function
                # factor 1000 because of delta function measured on time scale ms
                f1_delta += self.nu * F1p * (t>=self.d)*E_m_d
                f2_delta += self.nu * F2p * E
                
                # contributions due to continuous part of a
                f1_a += - self.alpha*self.nu**2 * F1p * ( eta * E + self.f_1_star_a(pk, t, self.d, self.sigma_a ) )
                f2_a += - self.alpha*self.nu**2 * F2p * ( eta * (E_m_d+E_p_d) + self.f_2_star_a(pk, t, self.d, self.sigma_a) )
            
            c =  np.real(f1_delta + f1_a) * m1_mat + np.real(f2_delta + f2_a) * m3_mat
            
                                        
            if t_prime < 0.0:
                return c.transpose()
            else:
                return c

        return c



    def cross_cov_components(self):
        """Calculates the cross covariance matrix for positive t"""

        # mean rate feedback
        L = self.epsilon*self.NE*self.w*(1. - self.gamma*self.g)

        print 'L = ', L

        # poles of resonant denominator
        pks = self.get_poles_(L, self.poles_max)
    
        m1_mat = self.epsilon * self.w * np.matrix([ [1.0, -self.g], [1.0, -self.g] ])

        m3_mat = self.epsilon**2 * self.NE * self.w**2 * (1. + self.gamma*self.g**2) * np.matrix( [ [1.0, 1.0], [1.0, 1.0] ] )
        
        def F1k(p):
            return 1./((1. + p*self.tau)*self.d + self.tau)

        def F2k_div_F1k(p):
            return 1. / (1. - p*self.tau - L*np.exp(p*self.d))

        eta = self.eta(self.sigma_a)
        
        def c_fb(t_prime):
            # use symmetry: c^T(t) = c(-t) to calculate for negative t
            if t_prime < 0:
                t = -t_prime
            else:
                t = t_prime


            # all poles are returned, conjugate pairs are in the list explicitly
            f1_delta = 0.0
            f1_a = 0.0
            
            for pk in pks:
                E = np.exp(pk*t)
                E_m_d = np.exp(pk*abs(t-self.d))
                F1p = F1k(pk)

                # contributions due to delta function
                # factor 1000 because of delta function measured on time scale ms
                f1_delta += self.nu * F1p * (t>=self.d)*E_m_d
                
                # ontributions due to continuous part of a
                f1_a += - self.alpha*self.nu**2 * F1p * ( eta * E + self.f_1_star_a(pk, t, self.d, self.sigma_a ) )
            
            c =  np.real(f1_delta + f1_a) * m1_mat
                                                    
            if t_prime < 0.0:
                return c.transpose()
            else:
                return c


        def c_ff(t_prime):
            # use symmetry: c^T(t) = c(-t) to calculate for negative t
            if t_prime < 0:
                t = -t_prime
            else:
                t = t_prime

            # all poles are returned, conjugate pairs are in the list explicitly
            f2_delta = 0.0
            f2_a = 0.0
            
            for pk in pks:
                E = np.exp(pk*t)
                E_m_d = np.exp(pk*abs(t-self.d))
                E_p_d = np.exp(pk*(t+self.d))
                F2p = F1k(pk) * F2k_div_F1k(pk)

                # contributions due to delta function
 
                f2_delta += self.nu * F2p * E
                
                # contributions due to continuous part of a
                f2_a += - self.alpha*self.nu**2 * F2p * ( eta * (E_m_d+E_p_d) + self.f_2_star_a(pk, t, self.d, self.sigma_a) )
            
            c =  np.real(f2_delta + f2_a) * m3_mat
                                                    
            if t_prime < 0.0:
                return c.transpose()
            else:
                return c

        return c_ff, c_fb



    def int_cov_matrix(self):
        """Calculates the temporal integral of the cross covariance matrix for all t"""

        # mean rate feedback
        L = self.epsilon*self.NE*self.w*(1. - self.gamma*self.g)

        print 'L = ', L
    
        m1_mat = self.epsilon * self.w  / (1.-L) * np.matrix([ [2., 1.-self.g], [1.-self.g, -2.*self.g] ])
        m3_mat = self.epsilon**2 * self.NE * self.w**2 * (1. + self.gamma*self.g**2) / (1.-L)**2 * np.matrix( [ [1.0, 1.0], [1.0, 1.0] ] )
        
        eta = self.eta(self.sigma_a)
                
        f1 = self.nu - self.alpha*self.nu**2 * ( eta + np.sqrt(2.*np.pi) * self.sigma_a - self.Gp(0.0, -self.d, self.sigma_a) )
        f2 = self.nu - self.alpha*self.nu**2 * ( 2.*eta + self.Gp(0.0, self.d, self.sigma_a) - self.Gp(0.0, -self.d, self.sigma_a) )
                
        # contributions due to continuous part of a
                
        c = f1 * m1_mat + f2 * m3_mat
                                                            
        return c


    def int_cov_matrix_corrtrans(self):
        """Calculates the temporal integral of the cross covariance matrix for all t"""

        # mean rate feedback
        L = self.epsilon*self.NE*self.w*(1. - self.gamma*self.g)

        print 'L = ', L
    
        m1_mat = self.epsilon * self.w  / (1.-L) * np.matrix([ [2., 1.-self.g], [1.-self.g, -2.*self.g] ])
        m3_mat = self.epsilon**2 * self.NE * self.w**2 * (1. + self.gamma*self.g**2) / (1.-L)**2 * np.matrix( [ [1.0, 1.0], [1.0, 1.0] ] )

        # integral of autocorrelation function
        # integral of delta peak - integral over correlation hole
        int_A = self.nu * (1.- np.sqrt(2.*np.pi)*self.sigma_a*self.nu)
                                               
        c = int_A * ( m1_mat + m3_mat )
                                                            
        return c
