# Moritz Helias, Sacha van Albada
# Nov 2014


import numpy as np
import numpy.linalg as la
import scipy.special as sps
from scipy.optimize import fsolve

# index convention
E = 0
I = 1
X = 2

EE = 0
EI = 1
II = 2
EX = 3
IX = 4


class ginzburg:


    def __init__(self, N, J, K, m_x, theta):
        self.N = N
        self.J = J
        self.K = K
        self.m_x = m_x
        self.theta = theta
        

    def mu_sigma_from_m_corr(self, m, c, mu_x, sigma_x):
        """calculate the mean and variance of the input to a neuron
        given the mean activity and covariances
        mu_x: mean of noise input
        sigma_x: std of noise input"""
        
        m = np.hstack((m, self.m_x))
        C = np.array([ [ c[EE], c[EI], c[EX] ],
                       [ c[EI], c[II], c[IX] ],
                       [ c[EX], c[IX], 0.    ] ])

        # vector of variances
        a = m*(1. - m)

        KJ = self.K*self.J
        KJ2 = self.K*self.J*self.J

        mu = np.dot(KJ,m) + mu_x

        # take into account correlations in the input to determine variance
        sigma2_m = np.dot(KJ2, a)
        sigma2_c = np.diag( np.dot( np.dot(KJ, C), KJ.T ))
        
        sigma = np.sqrt(sigma2_m + sigma2_c + sigma_x**2)

        return mu, sigma


    def susceptibility_corr(self, m, c, mu_x, sigma_x):

        mu, sigma = self.mu_sigma_from_m_corr(m, c, mu_x, sigma_x)

        S = 1./(np.sqrt(2.*np.pi)*sigma) * np.exp( -(mu-self.theta)**2 / (2.*sigma**2) )
	# assumes a (near-)Heaviside gain function
        return S


    def correlations(self, m, S):
        """solve the linear system of equations determining the correlations
        given we know the susceptibility"""
        W = ((self.J*self.K).T * S).T

        N = self.N

        a = m*(1.-m)
        c = np.zeros(5, dtype=float)

        pref = a[X]/self.N[X] / ( (2.-W[E,E])*(2.-W[I,I]) - W[E,I]*W[I,E] )    
        c[EX] = pref * ( (2. - W[I,I])*W[E,X] + W[E,I]*W[I,X] )
        c[IX] = pref * ( (2. - W[E,E])*W[I,X] + W[I,E]*W[E,X] )

        M = np.matrix([[1. - W[E,E], - W[E,I], 0.],
                       [-0.5*W[I,E], 1. - 0.5*(W[E,E]+W[I,I]), -0.5*W[E,I]],
                       [0., -W[I,E], 1. - W[I,I]]])

        b = np.array([W[E,E]*a[E]/N[E], 0.5*(W[I,E]*a[E]/N[E] + W[E,I]*a[I]/N[I]), W[I,I]*a[I]/N[I]])

        b1 = np.array( [W[E,X]*c[EX], 0.5*(W[E,X]*c[IX] + W[I,X]*c[EX]), W[I,X]*c[IX]] )

        c[:EX] = np.linalg.solve(M, b + b1)

        return c


    def solve_self_consistently_corr(self, m0, mu_x, sigma_x):
        """determine self-consistent rates and correlations
        taking effect of correlations on input statistics into account"""

        c = np.zeros(5)
        
        def m_zero_corr(m):

            mu, sigma = self.mu_sigma_from_m_corr(m, c, mu_x, sigma_x)

            new_m = 0.5*(1-sps.erf(-(mu-self.theta)/(np.sqrt(2.)*sigma)))

            return new_m - m

        err = 1.
        while err > 1e-15:
            m0 = fsolve(m_zero_corr, m0)
            S = self.susceptibility_corr(m0, c, mu_x, sigma_x)
            c_new = self.correlations(np.hstack((m0, self.m_x)), S)
            err = la.norm(c_new - c)
            c = c_new

        print 'error ', err

        return m0, c_new


    def corr_vs_lag(self, m0, tau, tlag, mu_x, sigma_x):
        """determine intrinsic average pairwise cross-correlations as a 
        function of time lag using eigendecomposition of connectivity matrix
        m0: starting guess for mean activities
        tlag should be in the form [min,max,step] and the result should include 0"""

        # need to calculate correlations to the largest absolute lag
        tlag_temp = np.arange(-max(abs(tlag[0:2])), max(abs(tlag[0:2]))+tlag[2], tlag[2])
        nlag_temp = len(tlag_temp)
        i_pos = np.min(np.where(tlag_temp>0))
        tlag_pos = tlag_temp[tlag_temp>0]
        nlag_pos = len(tlag_pos)

        i0 = np.where(abs(tlag_temp-tlag[0])<tlag[2]/2.)[0][0]
        i1 = np.where(abs(tlag_temp-tlag[1])<tlag[2]/2.)[0][0]

        # zero-lag correlations
        m, c0 = self.solve_self_consistently_corr(m0, mu_x, sigma_x)
        S = self.susceptibility_corr(m, c0, mu_x, sigma_x)
        W = ((self.J[:,0:2]*self.K[:,0:2]).T * S).T
        P = np.diag([1.,1.])-W
        A = np.diag(m[0:2]*(1.-m[0:2])/self.N[0:2])
        # right eigenvectors
        l, u = np.linalg.eig(P)
        # left eigenvectors
        l, v = np.linalg.eig(P.T)
        # normalization
        f = np.sqrt(np.dot(v[:,0], u[:,0].T))
        u = u/f
        v = v/f
        c = np.zeros([2, 2, nlag_temp])
        
        # Ginzburg & Sompolinsky (1994) Eq. 6.20 except, unlike that paper, 
        # we take positive lags to correspond to the target population 
        # following the source population. 

#        for j in xrange(2): # target population index
#            for k in xrange(2): # source population index
#                for i in xrange(2): # eigenvector index
#                    c[k,j,i_pos:] += (np.dot([[c0[EE], c0[EI]],[c0[EI], c0[II]]], v[:,i])[j] + \
#                    m[j]*(1.-m[j])/self.N[j]*v[j,i])*u[k,i]*np.exp(-tlag_pos/tau*l[i])
#                if j==k:
#                    c[k,j,i_pos:] -= m[j]*(1.-m[j])/self.N[j]*np.exp(-tlag_pos/tau)
                    
# own, equivalent form:            
        for i in xrange(2):
            for k in xrange(2):
                # solution for disjoint pairs of populations
                pref = 2.*np.dot(v[:,i].T, np.dot(A, v[:,k]))*(1./(l[i] + l[k]))
                temp = np.outer(pref*np.outer(u[:,i], u[:,k]), np.exp(-l[i]/tau*tlag_pos))
                mat = np.reshape(temp, [2,2,nlag_pos])         
                c[:,:,i_pos:] += mat
        c[:,:,i_pos:] -= np.reshape(np.outer(A, np.exp(-tlag_pos/tau)),[2,2,nlag_pos])
        
        c[:,:,0:i_pos-1] = c[:,:,i_pos:].transpose([1,0,2])[:,:,-1::-1] # permute the first two axes, then invert the time axis
        c[:,:,i_pos-1] = [[c0[EE], c0[EI]],[c0[EI], c0[II]]]

        c = c[:,:, i0:i1+1]

        return c

