# Sacha van Albada, Nov 2014

# Random recurrent network of inhibitory LIF neurons
# with exponential current-based synapses

import nest
import numpy
from parameters_inhibitory_LIF_network import *


def create_and_connect_correlometer(pop1, pop2):
    cd = nest.Create('correlation_detector')
    # start recording at tmax to avoid fall-off to the sides
    nest.SetStatus(cd, {'Tstart': tmax, 'delta_tau': tbin, 'tau_max': tmax})
    nest.SetDefaults('static_synapse', {'weight': 1.0, 'receptor_type': 0})
    nest.ConvergentConnect(pop1, cd, model='static_synapse')
    nest.SetDefaults('static_synapse', {'receptor_type': 1})
    nest.ConvergentConnect(pop2, cd, model='static_synapse')
    return cd


def write_autocorrelation_data(cds, fn):

    hs = numpy.zeros(2*tmax/tbin+1)
    n1 = n2 = 0
    
    for cd in cds:
        if nest.GetStatus(cd, 'local')[0]: 
            hs += nest.GetStatus(cd, 'histogram')[0]
            n1 += nest.GetStatus(cd, 'n_events')[0][0]
            n2 += nest.GetStatus(cd, 'n_events')[0][1]

    cdarr = numpy.array([n1, n2] + list(hs))
    numpy.savetxt(fn + str(nest.Rank()), cdarr)


def simulate(N0, N, scale_type, procs):
        
    K0 = int(epsilon*N0)

    if scale_type in ['vreeswijk', 'helias']:
        K = int(epsilon*N)

        # new synaptic strength
        if scale_type == 'vreeswijk':
            J = J0*numpy.sqrt(K0/float(K))
        if scale_type == 'helias':
            J = J0*K0/float(K)

    if scale_type == 'none':
        if N != N0:
            print 'warning: scale_type "none", so using N = ', N0
            N = N0
        K = K0
        J = J0

    # mean and variance of intrinsic inputs in this network (mV)
    mu_int = -tauMem/1000. * nu0 * K * J
    var_int = tauMem/1000. * nu0 * K * J**2

    # external drive (pA)
    mu_ext = CMem/tauMem * (mu - mu_int)
    sigma_ext = CMem/tauMem * (numpy.sqrt(sigma**2 - var_int))

    print 'N: ', N
    print 'K: ', K
    print 'J: ', J
    print 'mu_ext: ', mu_ext
    print 'sigma_ext: ', sigma_ext

    nest.ResetKernel()

    nest.SetKernelStatus({'resolution': dt, 'total_num_virtual_procs': procs, 'overwrite_files': True, 'print_time': False})
    nest.sr("/RandomConvergentConnect << /allow_multapses false >> SetOptions")


    print "Building network"

    neuron_params= {"C_m":        CMem,
                    "tau_m":      tauMem,
                    "t_ref":      tauRef,
                    "tau_syn_ex": tauSyn,
                    "tau_syn_in": tauSyn,
                    "E_L":        V_r,
                    "V_reset":    V_r,
                    "V_th":       theta,
                    "I_e": 	  mu_ext}

    nest.SetDefaults(nrn_model, neuron_params)

    nodes = nest.Create(nrn_model, N)

    noise = nest.Create("noise_generator")
    nest.SetStatus(noise, {"mean": 0., "std": sigma_ext})

    print "Connecting devices."

    nest.CopyModel("static_synapse", "inhibitory", {"weight":-J/tauSyn, "delay":delay})
 
    nest.DivergentConnect(noise, nodes, 1.0, delay)


    if N_rec_sd > 0:
        spikes=nest.Create("spike_detector")

        nest.SetStatus([spikes],[{"label": datapath + "brunel-py",
                                  "withtime": True,
                                  "withgid": True,
                                  "to_memory": False,
                                  "to_file": True}])

        nest.ConvergentConnect(range(1,1+N_rec_sd), spikes, model="excitatory")

    print "Connecting network."

    nest.RandomConvergentConnect(nodes, nodes, K, model="inhibitory")


    ############################################################
    ## connect recording neurons and correlation detectors
    ###########################################################

    # autocorrelations

    cd_acf = []

    for n in xrange(N_rec_acf):
        cd_acf.append( create_and_connect_correlometer([nodes[n]], [nodes[n]]) )

    # cross-correlations
    cd = create_and_connect_correlometer(nodes[:N_rec], nodes[N_rec:2*N_rec])

    print "Simulating."

    nest.Simulate(simtime)

    suffix = scale_type + '.dat'

    if nest.GetStatus(cd, 'local')[0]:
        hs = list(nest.GetStatus(cd, 'histogram')[0])
        n1 = nest.GetStatus(cd, 'n_events')[0][0]
        n2 = nest.GetStatus(cd, 'n_events')[0][1]
        cdarr = numpy.array([n1, n2] + hs)
        numpy.savetxt(datapath + 'corr_' + suffix, cdarr)

    write_autocorrelation_data(cd_acf, datapath + 'acf_' + suffix)

