# Sacha van Albada, Nov 2014

import sys
from params import *
import logging
import pyNN
import matplotlib.pyplot as plt
import time
import numpy

# prepare simulation
logging.basicConfig()
if simulator == 'nest' :
    import pyNN.nest as sim
sim.setup(**simulator_params[simulator])

if simulator == 'nest' :
    sim.nest.SetKernelStatus({'print_time' : False, 'dict_miss_is_error': False})

# to ensure pyNN.nest is imported before pyNN.random
import network

# create network
start_netw = time.time()
n = network.Network(sim)
n.setup(sim)
end_netw = time.time()
if sim.rank() == 0:
    print 'Creating the network took ', end_netw - start_netw, ' s'

# simulate
if sim.rank() == 0:
    print "Simulating..."
start_sim = time.time()
t = sim.run(simulator_params[simulator]['sim_duration'])
end_sim = time.time()
if sim.rank() == 0:
    print 'Simulation took ', end_sim - start_sim, ' s'

# write output
if sim.rank() == 0:
    # create directory if it does not yet exist
    if create_raster_plot or write_corr:
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

if write_spikes:
    if sim.rank() == 0:
        print "Writing spike data to file..."
    start_write = time.time()
    for layer in layers:
        for pop in pops:
            filename = output_dir + "/spikes_" + layer + pop + ".dat"
            n.pops[layer][pop].printSpikes(filename, gather=gather_output)
    end_write = time.time()
    if sim.rank() == 0:
        print "Writing spike data to file took ", end_write - start_write, " s"

if write_v:
    if sim.rank() == 0:
        print "Writing membrane potentials to file..."
    start_write = time.time()
    for layer in layers:
        for pop in pops:
            filename = output_dir + "/voltages_" + layer + pop + ".dat"
            n.pops[layer][pop].print_v(filename, gather=gather_output)
    end_write = time.time()
    if sim.rank() == 0:
        print "Writing membrane potentials to file took ", end_write - start_write, " s"

if print_stats or create_raster_plot or plot_corr or write_corr:
    spikes = n.get_spikes()

if print_stats or plot_corr or write_corr:
    rates = n.get_rates(spikes, t)

if print_stats:
    if record_v:
        v = {}
        for layer in numpy.sort(layers.keys()):
            v[layer] = {}
            for pop in pops:
                v[layer][pop] = n.pops[layer][pop].get_v(gather=True)

    if sim.rank() == 0:
        # Print mean rates
        print 'Rates:'
        for layer in numpy.sort(layers.keys()):
            for pop in pops:
                print layer + pop + ': ' + str(rates[layer][pop])


        # Print mean CVs of ISI distributions
        print 'CV_ISIs:'
        for layer in numpy.sort(layers.keys()):
            for pop in pops:
                if spikes[layer][pop] != []:
                    neurons = numpy.unique(spikes[layer][pop][:,0])
                else:
                    neurons = []
                    cv = []
                    for neuron in neurons:
                        intervals = numpy.diff(spikes[layer][pop][numpy.where(spikes[layer][pop][:,0]==neuron)[0],1])
                        if len(intervals):
                            cv.append(numpy.std(intervals)/numpy.mean(intervals))
                    print layer + pop + ': ' + str(numpy.mean(cv))

        if record_v:
        # Print mean membrane potentials
            print 'Mean membrane potentials:'
            for layer in numpy.sort(layers.keys()):
                for pop in pops:
                    if v[layer][pop] != []:
                        mean_v = numpy.mean(v[layer][pop][:,2])
                    else:
                        mean_v = numpy.nan
                    print layer + pop + ': ' + str(mean_v)


if create_raster_plot:
    colors = {'E': 'b', 'I': 'r'}
    offsets = [0]
    #plt.ion()
    for i, layer in enumerate(numpy.sort(layers.keys())):
        for j, pop in enumerate(pops):
            sp = spikes[layer][pop]
            which = [k for k in xrange(len(sp[:,1])) if sp[k,1] >= raster_t_min and sp[k,1] < raster_t_max]
            plt.plot(sp[which, 1], -(sp[which, 0] + offsets[i*n_pops_per_layer+j]), '.', color=colors[pop])
            if record_fraction:
                offsets.append(offsets[i*n_pops_per_layer+j] + round(len(n.pops[layer][pop])*frac_record_spikes))
            else:
                offsets.append(offsets[i*n_pops_per_layer+j] + n_record)
    plt.xlabel('t (ms)')
    temp = numpy.array(offsets)[0::2]
    ytick_pos=-temp[1:]+numpy.diff(temp)/2.
    ticklabels=[]
    for layer in numpy.sort(layers.keys()):
        ticklabels.append(layer)
    plt.yticks(ytick_pos, ticklabels)
    plt.ylim([-max(temp)-2,2])
    plt.savefig(output_dir + '/raster_plot.pdf')


if write_corr:
    # get covariance counts (sums of products of spike counts)
    if simulator == 'nest' :
        # the correlomatrix_detector only lies on one process
        if sim.nest.GetStatus(n.corr_detector, 'local')[0]:
            print 'getting count_covariance on rank ', sim.rank()
            cov = sim.nest.GetStatus(n.corr_detector, 'count_covariance')[0]
            delta_tau = sim.nest.GetStatus(n.corr_detector, 'delta_tau')[0]

            # compute normalized covariances between each pair of populations (Pearson correlation coefficients)
            cov_norm = {}
            # centered but without normalization, to prevent NaNs
            cov_center = {}
            # minimum normalized covariance across all populations and lags - for setting y-axis limits
            mincov_norm = 1
            for target_layer in numpy.sort(layers.keys()):
                for target_pop in pops:
                    target_index = structure[target_layer][target_pop]
                    cov_norm[target_index] = {}
                    cov_center[target_index] = {}
                    # convert to double to prevent overflow
                    n_events_target = numpy.double(sim.nest.GetStatus(n.corr_detector, 'n_events')[0][target_index])
                    # standard deviation times simulation length
                    target_sd = numpy.sqrt(cov[target_index][target_index][0] - n_events_target**2*delta_tau/(t-0.1-tau_max))
                    for source_layer in numpy.sort(layers.keys()):
                        for source_pop in pops:
                            source_index = structure[source_layer][source_pop]
                            n_events_source = numpy.double(sim.nest.GetStatus(n.corr_detector, 'n_events')[0][source_index])
                            source_sd = numpy.sqrt(cov[source_index][source_index][0] - n_events_source**2*delta_tau/(t-0.1-tau_max))
                            cov_center[target_index][source_index] = numpy.array(list(cov[target_index][source_index][::-1]) + list(cov[source_index][target_index][1:])) - n_events_target*n_events_source*delta_tau/(t-0.1-tau_max)
                            cov_norm[target_index][source_index] = cov_center[target_index][source_index]/(target_sd*source_sd)
                            if min(cov_norm[target_index][source_index]) < mincov_norm:
                                mincov_norm = min(cov_norm[target_index][source_index])

            if write_corr:
                f = open(output_dir + '/covariances.dat','w')
                print >>f, 'tau_max: ', tau_max
                print >>f, 'delta_tau: ', delta_tau, '\n'
                for target_layer in numpy.sort(layers.keys()):
                    for target_pop in pops:
                        target_index = structure[target_layer][target_pop]
                        for source_layer in numpy.sort(layers.keys()):
                            for source_pop in pops:
                                source_index = structure[source_layer][source_pop]
                                print >>f, target_layer, target_pop, '-', source_layer, source_pop
                                for i in xrange(len(cov_center[target_index][source_index])):
                                    print >>f, cov_center[target_index][source_index][i]
                                print >>f, ''
                f.close()

sim.end()
