# Sacha van Albada, Nov 2014

import numpy
import os
import sys

params_dict = {
'nest' : { 'time_scaling' : 1.,
           'parallel_safe' : False, # whether to make random numbers independent of the number of processes
           'gather_output' : True, # whether to gather spikes and voltages across processes
           'save_connections' : False,
           'output_dir' : "./data",
           'connect_type' : "multinomial", # "multinomial" uses the NEST developer function RandomPopulationConnectD
           'distribute_delays' : True, # whether to distribute delays
           'random_delays' : 'normal', # type of distribution for random delays if distribute_delays == True ('normal' or 'uniform'). For connect_type=multinomial, the delays are normally distributed regardless of this parameter.
           'scale_type' : 'w_and_ext', # for a detailed explanation, read below
           'w_and_g_equal_factor' : True, # whether to scale all excitatory weights by the same factor when scale_type = 'w_and_g
           'N_scaling' : 0.4,
           'K_scaling' : 0.4,
           'scale_ext' : False, # whether to downscale in-degrees for external synapses. Currently only implemented for scale_type w_and_ext and g_and_ext.
           'neuronmodel' : 'IF_curr_exp',
           'establish_connections' : True,
           'input_types' : ['poisson'], # Type of background input. Possible values: 'poisson', 'DC'. Leave empty for scale_type w_mean_var_ext_noise.
           'pool_poisson': False, # whether to draw Poisson input to each neuron from a pool of generators. Not yet implemented for nest_native_poisson=True.
           'n_poisson_generators': 1000, # total size of pool of Poisson generators if pool_poisson is True
           'poisson_rate': 40., # rate of each Poisson generator in spikes/s. If chosen as 800/n with n integer (for scale_ext=False) or 800*K_scaling/n (for scale_ext=True), the input rates of the model with independent Poisson generators can be precisely reproduced. A higher rate means that different neurons receive input from fewer common sources, but larger overlap between their spike trains in case of common sources.
           'nest_native_poisson': True, # When True, speeds up simulation by requiring only a single Poisson generator for each population
           'nest_native_noise': True, # for scale_type w_mean_var_ext_noise
           'DC_automatic' : False, # whether to set DC amplitude automatically to the mean standard Poisson input to each population
           'record_fraction' : False, # whether to record from a fixed fraction of neurons in each population. If False, a fixed number of neurons is recorded.
           'n_record' : 50, # Number of neurons from which to record spikes when record_fraction = False
           'frac_record_spikes' : 0.1, # Fraction of neurons from which to record spikes when record_fraction = True
           'write_spikes' : True, # whether to write spike data to file
           'record_v' : False, # Whether to record membrane potentials
           'write_v' : False, # Whether to write membrane potentials to file
           'n_record_v' : 20, # fixed number of neurons from which to record membrane potentials when record_v=True and record_fraction = False
           'frac_record_v' : 0.02, # Fraction of neurons from which to record membrane potentials if record_v=True and record_fraction = True
           'record_corr' : True, # Whether to record correlations - only for NEST back-end
           'write_corr'  : True, # Whether to write correlations to file
           'print_stats' : True, # whether to output summary statistics (mean rates, CV_ISIs, mean membrane potentials)
           'seeds' : numpy.random.randint(low=100000, high=10000000, size=24), # random number generator seeds for V and connectivity. When parallel_safe=True, only the first is used. When parallel_safe=False, the first num_processes are used.
           'master_seed' : 124678 # for NEST Poisson generators
    }
}

# simulator backend
simulator = 'nest'


# Load params from params_dict into global namespace
globals().update(params_dict[simulator])

# scale type. Possible values:

# 'w_and_ext' - scale in-degrees, synaptic weights, and external input to keep mean and variances of activity constant in the diffusion approximation

# 'w_and_g' - scale in-degrees, synaptic weights, and relative weights of inhibitory synapses to keep mean and variances of activity constant in the diffusion approximation. For this scale type, one can choose to scale all excitatory weights by the same factor, or make all excitatory weights equal after scaling.

# 'g_and_ext' - scale in-degrees, relative weights of inhibitory synapses, and external input to keep mean and variances of activity constant in the diffusion approximation

# 'w_mean_var_ext_poisson' - scale synaptic weights as well as the mean and variance of the external input following Helias et al. (2013) Eq. (E.1). This aims to preserve not only rates but also the shape of correlation functions (though not their scale unless the network size is constant). When downscaling, intrinsic variance increases and extrinsic variance needs to be reduced. This is only possible up to the point where the extrinsic variance vanishes. In this version, Poisson input is replaced by a combination of excitatory and inhibitory Poisson input, and any DC or noise input is left unchanged.

# 'w_mean_var_ext_noise' - As above, but adjusting the variance of the noise input, and leaving any DC or Poisson input unchanged.

# For the original full-scale model, use scale_type = None with N_scaling = 1.

# For the full-scale in-degrees, use scale_type = None with K_scaling = 1 and arbitrary N_scaling.

# With scale_type None, the in-degrees are nevertheless scaled by K_scaling

if neuronmodel == 'IF_curr_exp':
    g = -4.
    neuron_parameters = {'cm'        : 0.25, # nF
                         'i_offset'  : 0.0/time_scaling,
                         'tau_m'     : 10.0*time_scaling,
                         'tau_refrac': 2.0*time_scaling,
                         'tau_syn_E' : 0.5*time_scaling,
                         'tau_syn_I' : 0.5*time_scaling,
                         'v_reset'   : -65.0,
                         'v_rest'    : -65.0,
                         'v_thresh'  : -50.0}

simulator_params = {
  'nest' :
  {
    'timestep'        : 0.1*time_scaling,
    'threads'         : 1,
    'sim_duration'    : 30000.*time_scaling, # ms
    'min_delay'       : 0.1*time_scaling
  }
}


layers = {'L23':0, 'L4':1, 'L5':2, 'L6':3}
n_layers = len(layers)
pops = {'E':0, 'I':1}
n_pops_per_layer = len(pops)
structure = {'L23': {'E':0, 'I':1},
             'L4' : {'E':2, 'I':3},
             'L5' : {'E':4, 'I':5},
             'L6' : {'E':6, 'I':7}}

num_neurons = {
  'L23': {'E':20683, 'I':5834},
  'L4' : {'E':21915, 'I':5479},
  'L5' : {'E':4850,  'I':1065},
  'L6' : {'E':14395, 'I':2948}
}

establish_connections = True

#Probabilities for >=1 connection between neurons in the given populations. The first index is for the target population; the second for the source population
#             2/3e      2/3i    4e      4i      5e      5i      6e      6i
conn_probs = [[0.1009,  0.1689, 0.0437, 0.0818, 0.0323, 0.,     0.0076, 0.    ],
             [0.1346,   0.1371, 0.0316, 0.0515, 0.0755, 0.,     0.0042, 0.    ],
             [0.0077,   0.0059, 0.0497, 0.135,  0.0067, 0.0003, 0.0453, 0.    ],
             [0.0691,   0.0029, 0.0794, 0.1597, 0.0033, 0.,     0.1057, 0.    ],
             [0.1004,   0.0622, 0.0505, 0.0057, 0.0831, 0.3726, 0.0204, 0.    ],
             [0.0548,   0.0269, 0.0257, 0.0022, 0.06,   0.3158, 0.0086, 0.    ],
             [0.0156,   0.0066, 0.0211, 0.0166, 0.0572, 0.0197, 0.0396, 0.2252],
             [0.0364,   0.001,  0.0034, 0.0005, 0.0277, 0.008,  0.0658, 0.1443]]


K_ext = {
  'L23': {'E':1600, 'I':1500},
  'L4' : {'E':2100, 'I':1900},
  'L5' : {'E':2000, 'I':1900},
  'L6' : {'E':2900, 'I':2100}
}

# Mean rates in the full-scale model, necessary for scaling
# From 5 s NEST (SLI) simulation with RandomPopulationConnectD recording from 10% of neurons, 200 ms transient discarded:

full_mean_rates = {
  'L23': {'E':0.8529779174725983/time_scaling, 'I':2.7937392795883365/time_scaling},
  'L4' : {'E':4.389072721740454/time_scaling,  'I':5.9022699573430835/time_scaling},
  'L5' : {'E':7.681701030927835/time_scaling,  'I':9.282625786163523/time_scaling},
  'L6' : {'E':1.18137595552467/time_scaling,   'I':7.940051020408163/time_scaling}
}


V0_mean = -58.
V0_sd = 5.

# Background rate per synapse
bg_rate = 8./time_scaling # spikes/s

# DC amplitude if 'DC' is in 'input_type' and the DC input is not automatically set to the mean Poisson input. When 'DC' is not in 'input_type', the values are assumed to be zero (this is relevant for the g_and_ext and w_and_ext scaling options, where DC input may be present after scaling. When the DC amplitude is automatically set to the mean standard Poisson input, this is done before scaling, so the values may be adjusted.
DC_amp = {
  'L23': {'E':0., 'I':0.},
  'L4' : {'E':0., 'I':0.},
  'L5' : {'E':0., 'I':0.},
  'L6' : {'E':0., 'I':0.}
}

# SD of weight distribution as fraction of mean weight
w_rel = 0.1

if neuronmodel == 'IF_curr_exp':
    w_mean = 87.8e-3/time_scaling # nA
    w_ext = 87.8e-3/time_scaling # external weight
    # larger weight for L4E->L23E connections
    w_234 = 2*w_mean

# Means and standard deviations of delays (ms)
d_mean = {'E': 1.5*time_scaling, 'I': 0.75*time_scaling}
d_sd = {'E': 0.75*time_scaling, 'I': 0.375*time_scaling}

# Parameters for transient thalamic input
thalamic_input = False
thal_params = {
  'n_thal'      : 902,
  # connection probabilities
  'C'           : {'L23': {'E': 0, 'I':0},
                   'L4' : {'E':0.0983, 'I':0.0619},
                   'L5' : {'E': 0, 'I':0},
                   'L6' : {'E':0.0512, 'I':0.0196}},
  'rate'        : 15/time_scaling, # spikes/s;
  'start'       : 200.*time_scaling, # ms
  'duration'    : 10.*time_scaling # ms;
}


tau_max = 100.*time_scaling # Maximum delay over which to determine covariances
tau_max_plot = 30.*time_scaling # Maximum delay over which to plot covariances


# plotting parameters
create_raster_plot = False
raster_t_min = 0 # ms
raster_t_max = simulator_params[simulator]['sim_duration'] # ms

# make any changes to the parameters
if 'custom_params.py' in os.listdir('.'):
    execfile('custom_params.py')
