# Sacha van Albada
# Nov 2014

# Analysis tools for microcircuit model of Potjans & Diesmann (Cereb Cortex, 2014)
# Based in part on code by Tobias Potjans

import os
import numpy as np
import matplotlib.mlab as mlab
import matplotlib.pyplot as p
import csv
from scipy import stats
import pylab
import time

class data:
    def __init__(self, simtime, data_type, rec_type, spike_rec, vm_rec):
        self.spike_name  = 'spikes'
        self.vm_name     = 'voltages'
        self.conn_name   = '.conn'
        self.num_layers = 4
        self.num_pop_per_layer = [2, 2, 2, 2]
        self.default_pops = [['L23E', 'L23I'],['L4E', 'L4I'],['L5E', 'L5I'],['L6E', 'L6I']]
        self.default_pops_flat = ['L23E', 'L23I','L4E', 'L4I','L5E', 'L5I','L6E', 'L6I']
        self.pop_types = [['E', 'I'], ['E', 'I'],['E', 'I'],['E', 'I']]
        self.h = 1. # indicates that times are given in ms
        self.simtime = simtime
        self.rec_type = rec_type
        if rec_type == 'frac':
            # a fraction of neurons was recorded
            self.frac_record_spikes = spike_rec
            self.frac_record_vm = vm_rec
        if rec_type == 'abs':
            # an absolute number of neurons was recorded
            self.n_record_spikes = spike_rec
            self.n_record_vm = vm_rec
        self.set_fig_param_defaults()
        self.load_data(data_type)



    def load_data(self, data_type):
        files = os.listdir('.')
        files = np.sort(files)

        spike_files = []
        vm_files = []
        conn_files = []
        for f in files:
            if f.find(self.spike_name)>(-1): spike_files.append(f)
            if f.find(self.vm_name)>(-1): vm_files.append(f)
            if f.find(self.conn_name)>-1: conn_files.append(f)
        
        if 'spikes' in data_type:
            print 'loading spike_data'
            self.spike_data = {}
            self.popsize = {}
            self.nspikes = {}
            for f in spike_files:
                pop = f.split('_')[1].split('.')[0]
                if pop not in self.nspikes.keys():
                    self.nspikes[pop] = 0
                try:
                    new_data = np.loadtxt(f)
                except:
                    new_data = []
                    print 'Warning: data not loaded from ' + f
                    pass
                if pop in self.spike_data.keys():
                    if np.size(new_data)>0:
                        if len(self.spike_data[pop])>0:
                            if np.size(new_data) == 2:
                                if np.size(self.spike_data[pop])==2:
                                    self.spike_data[pop] = np.append([self.spike_data[pop]], [new_data], axis=0)
                                else:
                                    self.spike_data[pop] = np.append(self.spike_data[pop], [new_data], axis=0)
                            else:
                                self.spike_data[pop] = np.append(self.spike_data[pop], new_data, axis=0)
                        else:
                            self.spike_data[pop] = new_data
                else:
                    self.spike_data[pop] = new_data
                # load population sizes and numbers of spikes
                for line in open(f):
                    if pop not in self.popsize.keys():
                        if "size =" in line:
                            self.popsize[pop] = int(line.split('=')[1])
                    if "n =" in line:
                        self.nspikes[pop] += int(line.split('=')[1])
                        break
            # set num_neurons assuming default structure
            self.num_neurons = [[0,0],[0,0],[0,0],[0,0]]
            for i in xrange(4):
                for j in xrange(2):
                    if self.popsize.has_key(self.default_pops[i][j]):
                        self.num_neurons[i][j] = self.popsize[self.default_pops[i][j]]

  
        if 'vm' in data_type:          
            print 'loading vm_data'
            self.vm_data = {}
            for f in vm_files:
                pop = f.split('_')[1].split('.')[0]
                try:
                    new_data = np.loadtxt(f)
                except:
                    new_data = []
                    print 'Warning: data not loaded from ' + f
                    pass
                if pop in self.vm_data.keys():
                    if np.size(new_data)>0:
                        if len(self.vm_data[pop])>0:
                            if np.size(new_data) == 2:
                                if np.size(self.vm_data[pop])==2:
                                    self.vm_data[pop] = np.append([self.vm_data[pop]], [new_data], axis=0)
                                else:
                                    self.vm_data[pop] = np.append(self.vm_data[pop], [new_data], axis=0)
                            else:                                
                                self.vm_data[pop] = np.append(self.vm_data[pop], new_data, axis=0)
                        else:
                            self.vm_data[pop] = new_data
                else:
                    self.vm_data[pop] = new_data    


        if 'conn' in data_type:
            print 'loading connectivity data'
            self.conn_data = {}
            for f in conn_files:
                source_pop = int(f.split('population')[1][0])
                source_pop_name = self.default_pops_flat[source_pop]
                target_pop = int(f.split('population')[2][0])
                target_pop_name = self.default_pops_flat[target_pop]
                if source_pop_name not in self.conn_data:
                    self.conn_data[source_pop_name] = {}
                first = True
                with open(f, 'rb') as csvfile:
                    csvreader = csv.reader(csvfile, delimiter = ' ')
                    for row in csvreader:
                        if row[0] != '#' and row[0] != '':
                            rowdata = np.array(row[0].split('\t'), dtype=float)
                            if first:
                                data = [list(rowdata)]
                                first = False
                            else:
                                data = np.append(data, [list(rowdata)], axis=0) 
                if target_pop_name in self.conn_data[source_pop_name]:
                    self.conn_data[source_pop_name][target_pop_name] = np.append(self.conn_data[source_pop_name][target_pop_name], data, axis=0)
                else:
                    self.conn_data[source_pop_name][target_pop_name] = data
     
        
################################ Analysis functions

    def check_vm(self, dt):
        '''
        Check if vm data are complete 
        '''
        required_len = self.simtime/dt+1
        not_complete = {}
        neurons_missing = {}
        time_complete = True
        neurons_complete = True
        for pop in self.vm_data:
            if self.rec_type == 'frac':
                required_n_neurons = round(self.frac_record_vm*self.popsize[pop])
            if self.rec_type == 'abs':
                required_n_neurons = self.n_record_vm
            not_complete_pop = {}
            data = self.vm_data[pop]
            neurons = np.unique(data[:,1])
            if len(neurons) != required_n_neurons:
                neurons_complete = False
                neurons_missing[pop] = required_n_neurons - len(neurons)
            for neuron in neurons:
                neuron_data = data[data[:,1]==neuron]
                if len(neuron_data) != required_len:
                    time_complete = False
                    not_complete_pop[neuron] = len(neuron_data)
            not_complete[pop] = not_complete_pop
        if time_complete and neurons_complete:
            print "vm data are complete"
        else:
            if not time_complete:
                print 'neurons for which not all vm values are present:', not_complete
            if not neurons_complete:
                print 'numbers of neurons missing:', neurons_missing


    def create_pop_rates(self, tmin, tmax):
        self.pop_rates = {}
        #try:
        for pop in self.spike_data:
            if np.size(self.spike_data[pop])>2:
                spike_times = self.spike_data[pop][:,0]
            else:
                spike_times = self.spike_data[pop]
            n_spikes = len(self.spike_data[pop][np.all([spike_times >= tmin, spike_times < tmax], axis=0)])
            if self.rec_type == 'abs':
                self.pop_rates[pop] = 1000.*n_spikes/(self.n_record_spikes*(tmax-tmin))
            if self.rec_type == 'frac':
                self.pop_rates[pop] = 1000.*n_spikes/(round(self.popsize[pop]*self.frac_record_spikes)*(t_max-t_min))
        # create structured_pop_rates according to standard structure
        self.structured_pop_rates = [[0,0],[0,0],[0,0],[0,0]]
        for i in xrange(4):
            for j in xrange(2):
                self.structured_pop_rates[i][j] = self.pop_rates[self.default_pops[i][j]]


    def create_pop_cv_isi(self, tmin, tmax):
        self.pop_cv_isi = {}
        #try:
        for pop in self.spike_data:
            if np.size(self.spike_data[pop])>2:
                spike_times = self.spike_data[pop][:,0]
            else:
                spike_times = self.spike_data[pop]
            data = self.spike_data[pop][np.all([spike_times >= tmin, spike_times < tmax], axis=0)]
            cv_isi = []
            if len(data) >2:
                for i in np.unique(data[:,1]):
                    intervals = np.diff(data[mlab.find(data[:,1]==i),0])
                    if (len(intervals)>1):
                        cv_isi.append(np.std(intervals)/np.mean(intervals))
            self.pop_cv_isi[pop] = np.mean(cv_isi)
                
        # create structured_cv_isi according to standard structure
        self.structured_cv_isi = [[0,0],[0,0],[0,0],[0,0]]
        for i in xrange(4):
            for j in xrange(2):
                self.structured_cv_isi[i][j] = self.pop_cv_isi[self.default_pops[i][j]]


    def create_pop_sync(self, tmin, tmax, dt):
        self.pop_sync = {}
        for pop in self.spike_data:
            data = self.spike_data[pop]
            self.pop_sync[pop] = sync(data, dt, tmin, tmax, self.h)


    def create_psth(self, bin_width, *arguments):
        """psth for all layers of spike_data made available
        in psth.

        Parameters:
        - bin_width in [ms]
        - (optional) t_min and t_max in [ms]
        """
        
        h = self.h
        if (len(arguments)>0):
            t_min = arguments[0]
        else: t_min = 0
        if (len(arguments)>1):
            t_max = arguments[1]
        else:
            t_max = self.simtime
        self.set_metadata('psth',
                          {'bin_width': bin_width, 
                           't_min': t_min, 
                           't_max': t_max})
        self.psth = []
        for i in range(self.num_layers):
            layer_psth = []
            for j in range(self.num_pop_per_layer[i]):
                pop = self.default_pops[i][j]
                layer_psth.append(psth(self.spike_data[pop], 
                                       bin_width, 
                                       t_min, 
                                       t_max, 
                                       h))
            self.psth.append(layer_psth)
                
############################## Basic analysis and accessory plotting functions


def psth(data_array, bin_width, t_min, t_max, h):
    """Create psth of data_array between t_min and t_max
    and for bin_width.

    Arguments:
    - data_array: column 0: neuron_ids, column 1: spike times
    - t_min and t_max in [ms]
    - bin_width [ms]
    - h: transform times in ms to format used in data_array
    """
    p = np.zeros(len(np.arange(t_min, t_max, bin_width)))
    if np.size(data_array)>2:
        for event in data_array:
            if ((event[0] > t_min*h) and (event[0] < t_max*h)):
                bin = int(np.floor((event[0]-t_min*h)/(bin_width*h)))
                p[bin] = p[bin] + 1
    if np.size(data_array)==2 and len(data_array)==1: # of the form [[d1,d2]]
        bin = int(np.floor((data_array[0][0]-t_min*h)/(bin_width*h)))
        p[bin] = p[bin] + 1
    if np.size(data_array)==2 and len(data_array)==2: # of the form [d1,d2]
        print 'data_array:', data_array
        bin = int(np.floor((data_array[0]-t_min*h)/(bin_width*h)))
        p[bin] = p[bin] + 1
    return p


def sync(data_array, bin_width, t_min, t_max, h):
    """Compute synchrony as variance of the population signal 
       normalized by the mean variance of the individual neurons. 
       See e.g. Morrison et al. (2007) Neural Comput, Eq. (5.1)
       Only those neurons that spiked at least once are taken into account.
    """
    if np.size(data_array)>2:
        spikes = data_array[:,0]
        senders = data_array[:,1]
    else:
        spikes = data_array[0]
        senders = data_array[1]
    neurons = np.unique(senders)
    n_neurons = len(neurons)
    pop_signal = psth(data_array, bin_width, t_min, t_max, h)
    neuron_var = 0
    for ii in xrange(n_neurons):
        neuron_data = data_array[pylab.find(senders==neurons[ii])]
        neuron_signal = psth(neuron_data, bin_width, t_min, t_max, h)
        neuron_var += np.var(neuron_signal)
    return np.var(pop_signal)/(neuron_var*n_neurons)




