'''
plot_utils.py

A collection of utility functions/classes for flexible layout of panels and axes

Written by Junji Ito j.ito@fz-juelich.de on 2013/07/24
'''
import os
import datetime
import cPickle
import json

from mpl_toolkits.axes_grid1 import Grid, ImageGrid
from matplotlib import gridspec


class ParamFactory(object):
    def __init__(self, labels=None, panelgrid=None, rowspan=None, colspan=None, axeslayout=None, cbar_size=None, labelsize=15):
        if labels is None:
            return
        else:
            self.labels = labels
        
        if isinstance(panelgrid[0], int):
            self.panelgrid = [panelgrid for _ in labels]
        else:
            self.panelgrid = panelgrid
        
        if None in [rowspan, colspan]:
            if len(labels) != self.panelgrid[0][0] * self.panelgrid[0][1]:
                raise ValueError("The number of panels does not match the specified panelgrid.")
            i_label = 0
            self.rowspan = []
            self.colspan = []
            for i_row in range(self.panelgrid[0][0]):
                for i_col in range(self.panelgrid[0][1]):
                    self.rowspan.append([i_row, i_row + 1])
                    self.colspan.append([i_col, i_col + 1])
        else:
            self.rowspan = rowspan
            self.colspan = colspan
            
        if axeslayout is None:
            self.axeslayout = [[1,1] for _ in self.labels]
        else:
            self.axeslayout = axeslayout
        
        self.cbar_size = cbar_size
        self.labelsize = labelsize
        
        # define the list of keys in param.json file
        self.panelprop_keys = ['cbar_pad', 'cbrange', 'cbticks', 'cbticklabels', 'cblabel']
        self.axesprop_keys = ['xrange', 'yrange', 'xticks', 'yticks', 'xticklabels', 'yticklabels', 'xlabel', 'ylabel', 'grid', 'spine_locs', 'noshow']
        
        # store frequently used values
        N_panel = len(self.labels)
        N_axes = {}
        for i,label in enumerate(self.labels):
            N_axes[label] = self.axeslayout[i][0] * self.axeslayout[i][1]
            
        # define json templates
        self.panelprop_null_template = '\t"%(key)s": {\n'
        for label in labels[:-1]:
            self.panelprop_null_template += '\t\t"%(label)s": null,\n' % {'label': label,}
        self.panelprop_null_template += '\t\t"%(label)s": null\n' % {'label': labels[-1],}
        self.panelprop_null_template += '\t\t}'
        
        self.axesprop_null_template = '\t"%(key)s": {\n'
        for label in labels[:-1]:
            self.axesprop_null_template += '\t\t"%(label)s": [' % {'label': label,} + 'null, ' * (N_axes[label] - 1) + 'null],\n'
        self.axesprop_null_template += '\t\t"%(label)s": [' % {'label': labels[-1],} + 'null, ' * (N_axes[labels[-1]] - 1) + 'null]\n'
        self.axesprop_null_template += '\t\t}'
        
        # generate json as text
        self.gen_jsontxt()
        
    def gen_jsontxt(self):
        # store frequently used values
        N_panel = len(self.labels)
        N_axes = {}
        for i,label in enumerate(self.labels):
            N_axes[label] = self.axeslayout[i][0] * self.axeslayout[i][1]
            
        
        # generate json object as text
        json_txt = '{\n'
        
        ## put values of given panel parameters 
        json_txt += '\t"labels": ['
        for i in range(N_panel - 1):
            json_txt += '"%s", ' % (self.labels[i],)
        json_txt += '"%s"],\n\n' % (self.labels[-1])
        
        json_txt += self.panelprop_null_template % {'key': 'labelpos'}
        json_txt += ',\n\n'
        
        json_txt += '\t"labelsize": {\n'
        for i in range(N_panel - 1):
            json_txt += '\t\t"%s": %d,\n' % (self.labels[i], self.labelsize)
        json_txt += '\t\t"%s": %d\n' % (self.labels[-1], self.labelsize)
        json_txt += '\t\t},\n\n'
        
        json_txt += '\t"panelgrid": {\n'
        for i in range(N_panel - 1):
            json_txt += '\t\t"%s": [%d, %d],\n' % (self.labels[i], self.panelgrid[i][0], self.panelgrid[i][1])
        json_txt += '\t\t"%s": [%d, %d]\n' % (self.labels[-1], self.panelgrid[-1][0], self.panelgrid[-1][1])
        json_txt += '\t\t},\n\n'
        
        json_txt += '\t"rowspan": {\n'
        for i in range(N_panel - 1):
            json_txt += '\t\t"%s": [%d, %d],\n' % (self.labels[i], self.rowspan[i][0], self.rowspan[i][1])
        json_txt += '\t\t"%s": [%d, %d]\n' % (self.labels[-1], self.rowspan[-1][0], self.rowspan[-1][1])
        json_txt += '\t\t},\n\n'
        
        json_txt += '\t"colspan": {\n'
        for i in range(N_panel - 1):
            json_txt += '\t\t"%s": [%d, %d],\n' % (self.labels[i], self.colspan[i][0], self.colspan[i][1])
        json_txt += '\t\t"%s": [%d, %d]\n' % (self.labels[-1], self.colspan[-1][0], self.colspan[-1][1])
        json_txt += '\t\t},\n\n'
        
        json_txt += '\t"axeslayout": {\n'
        for i in range(N_panel - 1):
            json_txt += '\t\t"%s": [%d, %d],\n' % (self.labels[i], self.axeslayout[i][0], self.axeslayout[i][1])
        json_txt += '\t\t"%s": [%d, %d]\n' % (self.labels[-1], self.axeslayout[-1][0], self.axeslayout[-1][1])
        json_txt += '\t\t},\n\n'
        
        json_txt += self.panelprop_null_template % {'key': 'axes_pad'}
        json_txt += ',\n\n'
            
        if self.cbar_size:
            json_txt += '\t"cbar_size": {\n'
            for i in range(N_panel - 1):
                json_txt += '\t\t"%s": %s,\n' % (self.labels[i], self.cbar_size[i])
            json_txt += '\t\t"%s": %s\n' % (self.labels[-1], self.cbar_size[-1])
            json_txt += '\t\t},\n\n'
        else:
            json_txt += self.panelprop_null_template % {'key': 'cbar_size'}
            json_txt += ',\n\n'
            
        for key in self.panelprop_keys[:-1]:
            json_txt += self.panelprop_null_template % {'key': key}
            json_txt += ',\n\n'
        json_txt += self.panelprop_null_template % {'key': self.panelprop_keys[-1]}
        json_txt += ',\n\n'
            
        ## put null-placeholders for axes properties 
        for key in self.axesprop_keys[:-1]:
            json_txt += self.axesprop_null_template % {'key': key}
            json_txt += ',\n\n'
        json_txt += self.axesprop_null_template % {'key': self.axesprop_keys[-1]}
        json_txt += '\n'
        
        json_txt += '}'
        
        self.jsontxt = json_txt
        
    def get(self):
        return(json.loads(self.jsontxt))
    
    def save(self, filename, backup=True):
        # If there is already a file of the same name, it is backed up with another file name
        if backup:
            if os.path.isfile(filename):
                bakfilename = "%s.%s.bak" % (filename, datetime.datetime.now().strftime("%y%m%d-%H%M%S"))
                os.rename(filename, bakfilename)
                print "WARNING: existing file %s is backed-up as %s" % (filename, bakfilename)
        fh = open(filename, "w")
        fh.write(self.jsontxt)
        fh.close
        print "Created a plot parameter file %s" % (filename,)
    
    def fromfile(self, filename):
        fh = open(filename)
        jsondict = json.load(fh)
        fh.close()
        return jsondict


def setspines(axes, locs, position=('outward', 4)):
    if 'all' in locs:
        pass
    else:
        # first remove all spines and ticks
        axes.xaxis.set_ticks_position('none')
        axes.yaxis.set_ticks_position('none')
        for loc in ('left', 'right', 'bottom', 'top'):
            axes.spines[loc].set_color('none')
        
        # then set spines and ticks at specified locations
        for loc in locs:
            axes.spines[loc].set_color('black')
            axes.spines[loc].set_position(position)
            if loc in ('left', 'right'):
                axes.yaxis.set_ticks_position(loc)
            elif loc in ('top', 'bottom'):
                axes.xaxis.set_ticks_position(loc)

def genpanels(fig, param):
    panels = {}
    for label in param['labels']:
        # set axes in panel
        grid = param['panelgrid'][label]
        rowspan = param['rowspan'][label]
        colspan = param['colspan'][label]
        gs = gridspec.GridSpec(grid[0], grid[1])
        if param['cbar_size'][label]:
            panel = ImageGrid(fig, gs[rowspan[0]:rowspan[1], colspan[0]:colspan[1]], param['axeslayout'][label], aspect=False, cbar_mode='single', cbar_location='right', cbar_pad=param['cbar_pad'][label], cbar_size=param['cbar_size'][label])
        else:
            panel = Grid(fig, gs[rowspan[0]:rowspan[1], colspan[0]:colspan[1]], param['axeslayout'][label])
        panels[label] = panel
      
    return panels
 
def setparams(panels, param):
    for label in param['labels']:
        panel = panels[label]
        
        # set axes params
        if param['axes_pad'][label]:
            panel.set_axes_pad(param['axes_pad'][label])
            
        if param['cbar_size'][label]:
            axes = panel.cbar_axes[0]
            if param['cblabel'][label]:
                axes.set_ylabel(param['cblabel'][label], rotation=-90)
            if param['cbrange'][label]:
                axes.set_ylim(param['cbrange'][label])
            if param['cbticks'][label]:
                axes.set_yticks(param['cbticks'][label])
            if param['cbticklabels'][label]:
                axes.set_yticklabels(param['cbticklabels'][label])
            if param['cblabel'][label]:
                axes.set_ylabel(param['cblabel'][label])
            
        for n_axes, axes in enumerate(panel):
            # place spines
            if param['spine_locs'][label][n_axes]:
                setspines(panel[n_axes], param['spine_locs'][label][n_axes])
        
            # set axes range, ticks, and labels
            if param['xrange'][label][n_axes]:
                axes.set_xlim(param['xrange'][label][n_axes])
            if param['yrange'][label][n_axes]:
                axes.set_ylim(param['yrange'][label][n_axes])
            if param['xticks'][label][n_axes]:
                axes.set_xticks(param['xticks'][label][n_axes])
            if param['yticks'][label][n_axes]:
                axes.set_yticks(param['yticks'][label][n_axes])
            if param['xticklabels'][label][n_axes]:
                axes.set_xticklabels(param['xticklabels'][label][n_axes])
            if param['yticklabels'][label][n_axes]:
                axes.set_yticklabels(param['yticklabels'][label][n_axes])
            if param['xlabel'][label][n_axes]:
                axes.set_xlabel(param['xlabel'][label][n_axes])
            if param['ylabel'][label][n_axes]:
                axes.set_ylabel(param['ylabel'][label][n_axes])
                
            if param['grid'][label][n_axes]:
                axes.grid(param['grid'][label][n_axes])

            # don't draw axes
            if param['noshow'][label][n_axes]:
                axes.set_axis_off()

def draw_labels(panels, param):
    for label in param['labels']:
        if label[0] not in ['`', '!', '@', '#', '$', '%', '^', '&', '*']:
            ax = panels[label][0]
            if param['labelpos'][label] is None:
                xpos = ax.get_xlim()[0]
                ypos = ax.get_ylim()[1]
                # multiply ypos by 1.1 when running fig_scaling_framework.py
            else:
                xpos, ypos = param['labelpos'][label]
            ax.text(xpos, ypos, label, size=param['labelsize'][label], weight='bold')
            
def show_layout(param, figsize=None, adjust=None):
    fig = plt.figure(figsize=figsize)
    
    if adjust:
        fig.subplots_adjust(*adjust)
        
    panels = genpanels(fig, param)
    setparams(panels, param)
    draw_labels(panels, param)
        
    plt.show()
                
                
class PlotData(object):
    def __init__(self):
        self.datanum = 0
        self.type = []
        self.args = []
        self.kwargs = []
        
    def append(self, type, args, kwargs={}):
        self.type.append(type)
        self.args.append(args)
        self.kwargs.append(kwargs)
        self.datanum += 1
    
    def plot(self, ax, cax=None):
        arts = []
        for i, plottype in enumerate(self.type):
            if plottype in ['boxes', 'boxplot']:
                arts.append(self.boxesplot(ax, i))
            elif plottype in ['pcolor', 'pcolormesh']:
                arts.append(self.pcolorplot(ax, i, cax))
            else:
                command = "arts.append(ax.%s(*self.args[i], **self.kwargs[i]))" % plottype
                exec(command)
        return arts
    
    def boxesplot(self, ax, i):
        # store original x-axis parameters
        xrange = ax.get_xlim()
        xticks = ax.get_xticks()
        xlabel = ax.get_xlabel()
        
        # plot box
        art = ax.boxplot(*self.args[i], **self.kwargs[i])
        
        # fix x-axis corrupted by boxplot
        ax.set_xlim(xrange)
        ax.set_xticks(xticks)
        ax.set_xlabel(xlabel)
        
        return art
    
    def pcolorplot(self, ax, i, cax=None):
        art = ax.pcolormesh(*self.args[i], **self.kwargs[i])
        if cax:
            clabel = cax.get_ylabel()
            cax.colorbar(art)
            cax.set_ylabel(clabel)
        return art
    
class PlotDataSet(object):
    def __init__(self, param):
        self.data = {}
        for key, val in param['axeslayout'].items():
            self.data[key] = [PlotData() for _ in range(val[0] * val[1])]
    
    def plot(self, panels, param):
        for label in param['labels']:
            panel = panels[label]
            for i_ax, ax in enumerate(panel):
                # plot data
                plotdata = self.data[label][i_ax]
                if 'pcolor' in plotdata.type:
                    cbax = panel.cbar_axes[0]
                else:
                    cbax = None
                plotdata.plot(ax, panel.cbar_axes[0])

def datasave(filename, plotdataset, backup=True):
    ## If there is already a file of the same name, it is backed up with another file name
    if backup:
        if os.path.isfile(filename):
            bakfilename = "%s.%s.bak" % (filename, datetime.datetime.now().strftime("%y%m%d-%H%M%S"))
            os.rename(filename, bakfilename)
            print "WARNING: existing file %s is backed-up as %s" % (filename, bakfilename)
    fh = open(filename, 'w')
    cPickle.Pickler(fh).dump(plotdataset)
    fh.close()
    print "Created a plot data file %s" % (filename,)
        
def dataload(filename):
    fh = open(filename, "r")
    plotdataset = cPickle.Unpickler(fh).load()
    fh.close()
    return plotdataset
        

if __name__ == '__main__':
    import numpy as np
    import matplotlib.pyplot as plt
    
    # simplest usage
    labels = ['A', 'B', 'C', 'D']       # labels for panels
    panelgrid = [2,2]       # specification of panel grid in the form of [#rows, #columns]
    paramfact = ParamFactory(labels, panelgrid)     # initialize the parameter factory
    param = paramfact.get()     # generate a dictionary that contains layout parameters
    show_layout(param, figsize=[10, 6])     # utility function to check the layout by drawing axes only
    
    # flexible layout
    labels = ['A', 'B', 'C', 'D']
    panelgrid = [3,3]
    rowspan = [[0,1], [1,3], [1,2], [2,3]]      # extent of each panel can be specified in terms of 
    colspan = [[0,3], [0,1], [1,2], [1,3]]      # spans in row (vertical) and column (horizontal)
    param = ParamFactory(labels, panelgrid, rowspan, colspan).get()
    show_layout(param, figsize=[10, 6])
    
    # more flexible and direct layout using a dense grid
    labels = ['A', 'B', 'C', 'D']
    panelgrid = [600, 1000]
    rowspan = [[50,300], [100,350], [360,550], [420,550]]
    colspan = [[150,550], [620,950], [100,300], [400,900]]
    param = ParamFactory(labels, panelgrid, rowspan, colspan).get()
    show_layout(param, figsize=[10, 6], adjust=[0,0,1,1])
    # 'adjust' is the arguments to plt.subplots_adjust. Here the margins in
    # all edges are set 0, so that the grid coordinate spans the whole canvas
    
    # a panel can contain a grid of axes
    labels = ['A', 'B', 'C', 'D']
    panelgrid = [600, 1000]
    rowspan = [[50,300], [100,350], [360,550], [420,550]]
    colspan = [[150,550], [620,950], [100,300], [400,900]]
    axeslayout = [[2,1], [1,1], [1,1], [1,3]]       # layout of multiple axes within each panel can be specified in the form of [#rows, #columns]
    param = ParamFactory(labels, panelgrid, rowspan, colspan, axeslayout=axeslayout).get()
    show_layout(param, figsize=[10, 6], adjust=[0,0,1,1])
    
    # how to plot data in axes
    labels = ['A', 'B', 'C', '#C']      # Label of a panel that begins with '#' is not drawn in the plot
    panelgrid = [2,2]
    axeslayout = [[2,1], [1,1], [1,1], [1,3]]
    param = ParamFactory(labels, panelgrid, axeslayout=axeslayout).get()
    
    # plot commands
    fig = plt.figure()
    panels = genpanels(fig, param)      # 'panels' contain all axes in the figure
    panels['A'][0].plot(range(10), range(10))       # The i-th axes in panel X is referred to as 'panels['X'][i]'
    panels['B'][0].plot(range(10), range(10))
    panels['C'][0].plot(range(10), range(10))
    panels['#C'][1].plot(range(10), range(10))
    setparams(panels, param)        # parameters for axes (xrange, xlabel, and so on) are set by this command
    draw_labels(panels, param)      # panel-label is put to each panel by this command
    plt.show()      # there is no label for the fourth panel since its label begins with '#'
    
    # save and load parameters
    paramfact = ParamFactory(['A', 'B'], [1, 2], axeslayout=[[2,1], [1,1]], cbar_size=[0, 0.3])
    param = paramfact.get()
    show_layout(param, figsize=[10, 6])
    paramfact.save("~/test.json")     # parameters stored in the parameter factory are saved in a json file
    param_fromfile = ParamFactory().fromfile("~/test.json")       # json file is loaded
    print param == param_fromfile       # check the identity of the saved and the loaded parameters
    
    # arbitrary plot function
    def testplot(panels, param):
        x = np.linspace(-10, 10, 100)
        y = np.linspace(-10, 10, 100)
        X, Y = np.meshgrid(x, y)
        Z = X**2 + Y**2
        panels['A'][0].plot(x, 'b--')
        panels['A'][1].plot(y, 'r:')
        im = panels['B'][0].pcolormesh(X, Y, Z)
        panels['B'].cbar_axes[0].colorbar(im)
    fig = plt.figure(figsize=[10, 6])
    param = ParamFactory(['A', 'B'], [1, 2], axeslayout=[[2,1], [1,1]], cbar_size=[0, 0.3]).get()
    panels = genpanels(fig, param)
    testplot(panels, param)
    setparams(panels, param)
    draw_labels(panels, param)
    plt.show()
    
    # (modify "test.json" by hand, then) same plot again with a modified json file
    fig = plt.figure(figsize=[10, 6])
    param = ParamFactory().fromfile("~/test.json")
    panels = genpanels(fig, param)
    testplot(panels, param)
    setparams(panels, param)
    draw_labels(panels, param)
    plt.show()
    
    quit()
    
    
