# Moritz Helias, 2014

# Functions for determining firing rate and its derivative
# for LIF neurons with exponential current-based synapses

import scipy.special
import scipy.integrate
import numpy as np


# stationary firing rate of neuron with synaptic low-pass filter
# of time constant tau_s
# driven by Gaussian noise with mean mu and standard deviation sigma
# from Fourcaud & Brunel 2002
#
def nu0_fb(tau_m, tau_s, tau_r, V_th, V_r, mu, sigma):
    
    alpha = np.sqrt(2)*abs(scipy.special.zetac(0.5))

    # effective threshold    
    V_th1 = V_th + sigma*alpha/2.*np.sqrt(tau_s/tau_m)

    # effective reset    
    V_r1 = V_r + sigma*alpha/2.*np.sqrt(tau_s/tau_m)

    # use standard Siegert with modified threshold and reset
    return nu_0(tau_m, tau_r, V_th1, V_r1, mu, sigma)




# derivative of nu_0 w.r.t. input rate
# for low-pass filtered synapses with tau_s
# effective threshold and reset from Fourcaud & Brunel 2002
#
def d_nu_d_nu_in_fb(tau_m, tau_s, tau_r, V_th, V_r, mu, sigma, w):

    alpha = np.sqrt(2)*abs(scipy.special.zetac(0.5))

    y_th = (V_th - mu)/sigma
    y_r = (V_r - mu)/sigma

    y_th_fb = y_th + alpha/2.*np.sqrt(tau_s/tau_m)
    y_r_fb = y_r + alpha/2.*np.sqrt(tau_s/tau_m)

    nu0 = nu0_fb(tau_m, tau_s, tau_r, V_th, V_r, mu, sigma)


    # linear contribution
    lin = np.sqrt(np.pi) * (tau_m*nu0)**2 * w/sigma * (np.exp(y_th_fb**2) * (1 + scipy.special.erf(y_th_fb)) - np.exp(y_r_fb**2) * (1 + scipy.special.erf(y_r_fb)) )

    # quadratic contribution
    sqr = np.sqrt(np.pi) * (tau_m*nu0)**2 * w/sigma * (np.exp(y_th_fb**2) * (1 + scipy.special.erf(y_th_fb)) * 0.5*y_th*w/sigma - np.exp(y_r_fb**2) * (1 + scipy.special.erf(y_r_fb)) * 0.5*y_r*w/sigma )

    return lin + sqr, lin, sqr


#
# for mu < V_th
#
def siegert1(tau_m, tau_r, V_th, V_r, mu, sigma):

    y_th = (V_th - mu)/sigma
    y_r = (V_r - mu)/sigma
   
    def integrand(u):
        if u == 0:
            return np.exp(-y_th**2)*2*(y_th - y_r)
        else:
            return np.exp(-(u-y_th)**2) * ( 1.0 - np.exp(2*(y_r-y_th)*u) ) / u

    lower_bound = y_th
    err_dn = 1.0
    while err_dn > 1e-12 and lower_bound > 1e-16:
        err_dn = integrand(lower_bound)
        if err_dn > 1e-12:            
            lower_bound /= 2

    upper_bound = y_th
    err_up = 1.0
    while err_up > 1e-12:
       err_up = integrand(upper_bound)
       if err_up > 1e-12:
           upper_bound *= 2

    err = max(err_up, err_dn)

    return 1.0/(tau_r + np.exp(y_th**2)*scipy.integrate.quad(integrand, lower_bound, upper_bound)[0] * tau_m)
 
#
# for mu > V_th
#
def siegert2(tau_m, tau_r, V_th, V_r, mu, sigma):

    y_th = (V_th - mu)/sigma
    y_r = (V_r - mu)/sigma

    def integrand(u):
        if u == 0:
            return 2*(y_th - y_r)
        else:
            return ( np.exp(2*y_th*u -u**2) - np.exp(2*y_r*u -u**2) ) / u

    upper_bound = 1.0
    err = 1.0
    while err > 1e-12:
        err = integrand(upper_bound)
        upper_bound *= 2

    return 1.0/(tau_r + scipy.integrate.quad(integrand, 0.0, upper_bound)[0] * tau_m)



def nu_0(tau_m, tau_r, V_th, V_r, mu, sigma):
    """tau_m: membrane time constant"""
    """tau_r: refractory time constant"""
    """V_th: threshold"""
    """V_r: reset potential"""
    """mu: mean input"""
    """sigma: std of equivalent GWN input"""

    if mu <= V_th*0.95:
        return siegert1(tau_m, tau_r, V_th, V_r, mu, sigma)
    else:
        return siegert2(tau_m, tau_r, V_th, V_r, mu, sigma)


